from mathfield.api import *
from mathfield.models import *
from mathfield.widgets import *


__title__ = 'django-mathfield'
__version__ = '0.1.0'
__author__ = 'Jonathan Goodnow'
__license__ = 'BSD License'
__copyright__ = 'Copyright 2014 Jonathan Goodnow'


from __future__ import unicode_literals
from django import forms
from django.utils.safestring import mark_safe
from django.conf import settings
import textwrap
import json
import cgi

class MathFieldWidget(forms.Textarea):
    
    def render(self, name, value, attrs=None):
        output = super(MathFieldWidget, self).render(name, value, attrs)
        output = '<div id="{0}-container"><span>{1}</span></div>'.format(
            attrs['id'], output)

        if value:
            if isinstance(value, dict):
                raw = value['raw'] if 'raw' in value else ''
                html = value['html'] if 'html' in value else ''
                html = html.replace('"', '\\"').replace("'", "\\'")
            elif isinstance(value, basestring):
                raw = value
                html = ''
            raw = cgi.escape(raw.replace('\\', '\\\\'))
        else:
            raw = html = ''

        if hasattr(settings, 'STATIC_URL'):
            static_url = getattr(settings, 'STATIC_URL', {})
        else:
            static_url = '/static/'
        output += textwrap.dedent("""
            <link rel="stylesheet" type="text/css" 
                href="{static}mathfield/css/mathfield.css"/>
            <script type="text/javascript" 
                src="{static}mathfield/js/mathfield.min.js"></script>
            <script type="text/javascript">
                renderMathFieldForm("{id}", "{raw}", "{html}");
            </script>
        """.format(static=static_url, id=attrs['id'], raw=raw, html=html))

        return mark_safe(output)


# -*- coding: utf-8 -*-

import subprocess
import os
import re
import cgi
from django.utils.encoding import smart_unicode

# These are functions for developers to use externally

def store_math(raw='', html=''):
    """ MathFields must be stored in the database as a string containing both
        the raw math and html.

        Arguments:

        * raw: this is your raw math as either LaTeX or just regular text

        * html: if you already know the html, there's no sense in calculating it
                again! But if you don't know it, leave this one blank and it 
                will be done for you.

        Returns:

        * dictionary with keys 'raw' and 'html'

        NOTICES:

        * Blocks of math must be enclosed in dollar signs. 
          If you need the normal dollar signs, use dollar signs preceeded by 
          backslashes, `\$`.

        * NODE.JS MUST BE INSTALLED FOR THIS FUNCTION TO RUN
    """
    if raw == '' or html != '':
        return {'raw': raw, 'html': html}

    return {'raw': raw, 'html': render_to_html(raw)}


def render_to_html(raw):
    """ Returns the HTML of a string of LaTeX. It can also contain plain text.
        LaTeX must be surrounded by dollar signs. Preceed dollar signs with
        backslashes to write literal dollar signs.

        NODE.JS MUST BE INSTALLED FOR THIS FUNCTION TO RUN
    """
    if not raw:
        return ''

    reg = re.finditer(r"(^|(?<!\\))\$(([^\$]|\\\$)*[^\\])\$", raw)

    # generate_html.js must be passed all the math text ask command line args. 
    # The dollar signs get stripped in advanced because the shell will interpret 
    # those as variables. The program will return each math object separated by
    # newlines. KaTeX doesn't understand actual dollar signs if they are
    # followed by another character (like x=\$2), so add a space after those
    results = [(mat.start(2), 
                mat.end(2), 
                mat.group(2).strip().replace('\\$', '\\$ ')
              ) for mat in reg if mat]

    if results == []:
        return raw

    math_start_positions, math_end_positions, raw_math = zip(*results)

    # prepare the shell to get the LaTeX via a call to Node.js
    # the shell is not explicitly called so there's no danger of shell injection
    # The command `node` must be on the system path
    env = dict(os.environ)
    env['LC_ALL'] = 'en_US.UTF-8' # accept unicode characters as output
    try:
        p = subprocess.Popen([
            'node', 
            os.path.join(os.path.dirname(__file__), 'generate_html.js')] 
                + list(raw_math),
            env=env, 
            stdout=subprocess.PIPE, 
            stderr=subprocess.PIPE)
    except (WindowsError, OSError):
        raise NodeError("Node.js is not on your system path.")
    else:
        node_output, node_error = p.communicate()
    
    if node_error:
        raise NodeError(node_error)

    html_bits = node_output.strip('\n').split('\n')

    final = []
    loc = 0
    for index, code in enumerate(html_bits):
        # measurements are one off from the index of the math to eliminate the
        # dollar sign specifiers
        # KaTeX will handle HTML encoding for the math text, but regular text
        # must have HTML stripped out for security reasons.
        final.append(cgi.escape(raw[loc:math_start_positions[index]]
                        .strip('$').replace('\\$', '$')))
        final.append(smart_unicode(code))
        loc = math_end_positions[index] + 1

    final.append(cgi.escape(raw[loc:].replace('\\$', '$')))
    return u''.join(final)


class NodeError(Exception):
    """ Exception that gets raised when Node.js is not installed, or when
        it throws a runtime error.
    """
    pass


# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from django.db import models
from django.core import exceptions
from mathfield.api import store_math
import json


MathFieldValidationError = lambda self, value: exceptions.ValidationError(
    'Could not resolve "{0}" to a dictionary with only keys "raw" and "html"'
    .format(str(value)))


class MathField(models.TextField):

    description = 'Field that allows you to write LaTeX and display it as HTML.'

    __metaclass__ = models.SubfieldBase

    def to_python(self, value):
        """ The data is serialized as JSON with the keys `raw` and `html` where
            `raw` is the entered string with LaTeX and `html` is the html string
            generated by KaTeX. If this function gets just a string,
            then we need to generate the LaTeX.

            WARNING: Generating the LaTeX server-side requires Node.js to be
            installed. To generate the LaTeX client-side, make sure that you
            specify that the MathFields that you use are assigned to the widget
            `MathFieldWidget` in your app's admin.py.
        """
        if value is None:
            return None

        if value == "":
            return {'raw': '', 'html': ''}

        if isinstance(value, basestring):
            try:
                return dict(json.loads(value))
            except (ValueError, TypeError):
                # the value was stored as just a string. Try to compile it to
                # LaTeX and return a dictionary, or raise a NodeError
                return store_math(value)
        
        if isinstance(value, dict):
            return value
        
        return {'raw': '', 'html': ''}
        
    def get_prep_value(self, value):
        if not value:
            return json.dumps({'raw': '', 'html': ''})

        if isinstance(value, basestring):
            try:
                dictval = json.loads(value)
            except (ValueError, TypeError):
                # This means the user tried to pass just a string of text in.
                # The HTML will be generated manually, but this will only work
                # if node.js is installed on the server. Otherwise, a NodeError
                # will be raised.
                return json.dumps(store_math(value))
            else:
                if {'raw', 'html'} == set(dictval.keys()):
                    return value
                else:                
                    raise MathFieldValidationError(self, value)

        if isinstance(value, dict):
            if {'raw', 'html'} == set(value.keys()):
                return json.dumps(value)
            else:
                raise MathFieldValidationError(self, value)

        return json.dumps({'raw': '', 'html': ''})

    def formfield(self, **kwargs):
        defaults = {
            'help_text': ('Type text as you would normally, or write LaTeX ' 
                'by surrounding it with $ characters.')
        }
        defaults.update(kwargs)
        field = super(MathField, self).formfield(**defaults)
        field.max_length = self.max_length
        return field


#! /usr/bin/env python

import codecs
import os
from setuptools import setup

setup(
	name="django-mathfield",
	version="0.1.0",
	description="A Django model field for writing and displaying LaTeX",
	long_description=codecs.open(
		os.path.join(os.path.dirname(__file__), 'README.rst'), 'r', 'utf-8').read(),
	author="Jonathan Goodnow",
	author_email="jon@goodnow.io",
	url="https://github.com/jongoodnow/django-mathfield",
	keywords=["django", "math", "latex", "katex"],
	license="BSD",
	classifiers=[
		"Development Status :: 3 - Alpha",
		"Intended Audience :: Developers",
		"License :: OSI Approved :: BSD License",
		"Natural Language :: English",
		"Operating System :: OS Independent",
		"Programming Language :: Python :: 2.7",
		"Environment :: Web Environment",
		"Framework :: Django",
		"Topic :: Utilities",
	],
	packages=["mathfield"],
	test_suite="tests",
	include_package_data=True,
)